package piglets;

import universum.bi.*;

/**
 * Created by IntelliJ IDEA.
 * User: Tanya
 * Date: 02.01.2007
 * Time: 21:06:16
 */
public abstract class SuperPiglet implements Being, Comparable<SuperPiglet> {
    protected float     M;          // mass
    protected float     S;          // speed
    public Integer   ID;         // piglet's id

    public float getM() {
        return M;
    }

    protected DrawWell  ownedWell;

    protected abstract SuperHerder getHerder();

    public String getName() {
        return "Piglet";
    }

    public String getOwnerName() {
        return "cimbaluk";
    }

    private static final float [] search_dx = new float[] { 0, 1, 0, 1, 0, -1, -1 };
    private static final float [] search_dy = new float[] { 0, 0, 1, 1, -1, 0, -1 };

    public static float distance(Location from, Location to) {
        assert search_dx.length == search_dy.length;

        float x0 = from.getX();
        float y0 = from.getY();
        float x1 = to.getX();
        float y1 = to.getY();

        float minDistance = SuperHerder.INF;

        for (int i=0; i<search_dx.length; ++i) {
            float x = x1 + Constants.getWidth()  * search_dx[i];
            float y = y1 + Constants.getHeight() * search_dy[i];

            float dist = Math.max(Math.abs(x - x0), Math.abs(y - y0));
            minDistance = Math.min(minDistance, dist);
        }
        
        return minDistance;
    }

    public Location findStep(Location from, Location to) {
        float dist = distance(from, to);

        if (dist < SuperHerder.EPS) {
            return null;
        }

        if (dist < S + SuperHerder.EPS) {
            return to;
        }

        float x0 = from.getX();
        float y0 = from.getY();
        float x1 = to.getX();
        float y1 = to.getY();

        int minIndex = -1;
        float minDistance = SuperHerder.INF;

        for (int i=0; i<search_dx.length; ++i) {
            float x = x1 + Constants.getWidth()  * search_dx[i];
            float y = y1 + Constants.getHeight() * search_dy[i];

            dist = Math.max(Math.abs(x - x0), Math.abs(y - y0));
            if (dist < minDistance) {
                minDistance = dist;
                minIndex = i;
            }
        }

        assert minIndex != -1;
        x1 += Constants.getWidth()  * search_dx[minIndex];
        y1 += Constants.getHeight() * search_dy[minIndex];

        float x_ar = x1 - x0;
        float y_ar = y1 - y0;

        if (Math.abs(x_ar) > S) {
            x_ar = S * Math.signum(x_ar);
        }
        if (Math.abs(y_ar) > S) {
            y_ar = S * Math.signum(y_ar);
        }

        return new Location(from.getX() + (int)x_ar, from.getY() + (int)y_ar);
    }

    public void processEvent(BeingInterface bi, Event event) {
        switch (event.kind()) {
            case BEING_BORN:
                BeingParams bp = (BeingParams)event.param();
                M = bp.M;
                S = bp.S;
                ID = bi.getId(this);
                getHerder().register(this);
                break;
            case BEING_DEAD:
                getHerder().unregister(this);
                break;
        }            
    }

    public static int normalizeY(int y) {
        return (y + Constants.getHeight()) % Constants.getHeight();
    }

    public static int normalizeX(int x) {
        return (x + Constants.getWidth()) % Constants.getWidth();
    }

    public int compareTo(SuperPiglet o) {
        return ID.compareTo(o.ID);
    }

    public void reinit(UserGameInfo ugi) {
        getHerder().init(ugi);
//        getHerder().MAX_TURNS = Math.max(getHerder().MAX_TURNS, ugi.maxTurns);
    }
}
