package piglets.delta;

import piglets.SuperPiglet;
import piglets.SuperHerder;
import piglets.DrawWell;
import piglets.GladeCell;
import universum.bi.*;
import universum.util.Util;

/**
 * Created by IntelliJ IDEA.
 * User: Tanya
 * Date: 03.02.2007
 * Time: 12:20:57
 */

enum PigletState {
    EXPLORE,
    EATING,
    FEEDING,
    SURVIVE,
    ATTACK,
}

class EventException extends Exception {
    public final Event event;

    EventException(EventKind kind, Object arg) {
        this.event = new Event(kind, arg);
    }

    public EventException() {
        this.event = null;
    }
}

public class Piglet extends SuperPiglet {
    public static final float M0                    = 5;
    public static final float S0                    = 3;

    public static final float C_crit                = Constants.K_emin + 5 * Constants.K_masscost;
    public static final float C_norm                = C_crit + 5 * Constants.K_masscost;
    public static final float C_growt_weight        = 5.f;
    public static final float C_energy_weight       = 0.05f;
    public static final float C_steps_to_born       = 0.3f / Constants.K_bite;
    public static final float C_max_steps_to_dw     = 2.f;

    public static final int C_growth_barrier        = 30;
    public static final int C_fight_barrier         = 1000;
    public static final int C_enemy_history         = 1;

    private PigletState state = PigletState.EXPLORE;

    private float E;
    private Location location;
    private PointInfo pointInfo;
    private BeingInterface beingInterface;

    public Event makeTurn(BeingInterface bi) {
        herder.updateTurn(bi);

        // update piglet params
        beingInterface = bi;
        E = bi.getEnergy(this);
        location = bi.getLocation(this);
        pointInfo = bi.getPointInfo(this);

        // exec automaton
        final int MAX_STEPS = 100;
        for (int step_ = 0; step_ < MAX_STEPS; ++step_) {
            PigletState prevState = state;

            try {
                switch (state) {
                    case EXPLORE:
                        if (x1_1()) {
                            log("EXPLORE: x1_1 -> ATTACK");
                            state = PigletState.ATTACK;
                            break;
                        }

                        if (x1_2()) {
                            log("EXPLORE: x1_2 : z1");
                            z1();
                            break;
                        }

                        if (x1_3()) {
                            log("EXPLORE: x1_3 -> EATING");
                            state = PigletState.EATING;
                            break;
                        }

                        if (x1_4()) {
                            log("EXPLORE: x1_4 -> FEEDING");
                            state = PigletState.FEEDING;
                            break;
                        }

                        if (x1_5()) {
                            log("EXPLORE: x1_5 -> SURVIVE");
                            state = PigletState.SURVIVE;
                            break;
                        }

                        if (x1_6() || x1_7()) {
                            log("EXPLORE: x1_6 | x1_7 : z2");
                            z2();
                            break;
                        }

                        log("EXPLORE: otherwise -> SURVIVE");
                        state = PigletState.SURVIVE;
                        break;
                    
                    case EATING:
                        if (x2_1()) {
                            log("EATING: x2_1 : z2");
                            z2();
                            break;
                        }

                        if (x2_2()) {
                            log("EATING: x2_2 : z1");
                            z1();
                            break;
                        }

                        if (x2_3()) {
                            log("EATING: x2_3 -> SURVIVE");
                            state = PigletState.SURVIVE;
                            break;
                        }

                        log("EATING: otherwise : z3");
                        z3();
                        break;
                    
                    case FEEDING:
                        if (x2_1()) {
                            log("FEEDING: x2_1 : z2");
                            z2();
                            break;
                        }

                        if (x2_2()) {
                            log("FEEDING: x2_2 : z1");
                            z1();
                            break;
                        }

                        if (x3_1()) {
                            log("FEEDING: x3_1 -> EXPLORE");
                            state = PigletState.EXPLORE;
                            break;
                        }

                        log("FEEDING: otherwise : z3");
                        z3();
                        break;
                    
                    case SURVIVE:
                        if (x4_1()) {
                            log("SURVIVE: x4_1 -> EXPLORE");
                            state = PigletState.EXPLORE;
                            break;
                        }

                        if (x4_2()) {
                            log("SURVIVE: x4_2 : z4");
                            z4();
                            break;
                        }

                        log("SURVIVE: otherwise : z5");
                        z5();
                        break;
                    
                    case ATTACK:
                        if (x5_2()) {
                            log("ATTACK: x5_2 : z7");
                            z7();
                            break;
                        }

                        if (x5_1()) {
                            log("ATTACK: x5_1 : z6");
                            z6();
                            break;
                        }

                        log("ATTACK: otherwise -> EXPLORE");
                        state = PigletState.EXPLORE;
                        break;
                }
            } catch (EventException event) {
                return event.event;
            }

            // break cycle if no action occurs
            if (state == prevState) {
                break;
            }
        }

        return null;
    }

    private boolean x1_1() {
       Location enemyPos = findClosestEnemy();
       float enemyDist = enemyPos != null ? distance(location, enemyPos) : Herder.INF;
       return enemyDist < S + Herder.EPS;
    }

    private boolean x1_2() {
        return herder.hasEatingDrawWells() &&
            E > M * Constants.K_toborn &&
            (herder.MAX_TURNS - herder.getTurn()) * M * Constants.K_masscost >
            E * (0.5f - Constants.K_borncost - C_crit);
    }

    private boolean x1_3() {
        DrawWell drawWell = findClosestDrawWell();

        if (drawWell == null) {
            return false;
        }
        
        eatingDrawWell = drawWell;
        foundDestination = drawWell.pos;

        herder.lockWrite();
        drawWell.owners.add(this);
        herder.addEatingDrawWells(true);
        herder.unlockWrite();
        return true;
    }

    private boolean x1_4() {
        Location next = findClosestFeeding();
        if (next != null) {
            foundDestination = next;
            return true;
        }

        return false;
    }

    private boolean x1_5() {
        return E - M * C_crit < Constants.K_masscost * M * (herder.MAX_TURNS - herder.getTurn());
    }

    private Location foundDestination;
    private boolean x1_6() {
        foundDestination = findExploreStep();
        return foundDestination != null;
    }

    private boolean x1_7() {
        foundDestination = findClosestUnexploredCell();
        return foundDestination != null && checkDistanceConstraints(foundDestination);
    }

    private boolean x2_1() {
        return distance(location, foundDestination) > Herder.EPS;
    }

    private boolean x2_2() {
        return E > M * Constants.K_toborn;
    }

    private boolean x2_3() {
        final float canEat = eatingDrawWell.getCell().count / eatingDrawWell.owners.size();
        if (eatingDrawWell.owners.size() > 1 && canEat < (M - E) / C_steps_to_born) {
            herder.lockWrite();
            eatingDrawWell.owners.remove(this);
            herder.unlockWrite();

            herder.addEatingDrawWells(false);
            eatingDrawWell = null;

            return true;
        }
        return false;
    }

    private boolean x3_1() {
        final float count = pointInfo.getCount(this);
        final float growth = pointInfo.getGrowthRate(this);
        return count < Herder.EPS || isGrowingPeriod() && (count + growth * C_steps_to_born < M * Constants.K_toborn - E);
    }

    private boolean tryNotToSurvive = false;
    private boolean x4_1() {
        return tryNotToSurvive = !tryNotToSurvive;
    }

    private boolean x4_2() {
        return E > M * C_crit;
    }

    private boolean x5_1() {
        foundDestination = findClosestEnemy();
        float dist = foundDestination != null ? distance(location, foundDestination) : Herder.INF;

        return foundDestination != null && dist < S + Herder.EPS && dist > Herder.EPS;
    }

    private boolean x5_2() {
        foundDestination = findClosestEnemy();
        float dist = foundDestination != null ? distance(location, foundDestination) : Herder.INF;

        return foundDestination != null && dist < S + Herder.EPS && dist < Herder.EPS;
    }

    private void z1() throws EventException {
        throw new EventException(EventKind.ACTION_BORN, getParams());
    }

    private void z2() throws EventException {
        throw new EventException(EventKind.ACTION_MOVE_TO, findStep(location, foundDestination));
    }

    private DrawWell eatingDrawWell;
    private void z3() throws EventException {
        if (eatingDrawWell == null) {
            throw new EventException(EventKind.ACTION_EAT, M - E);
        }

        final float canEat = eatingDrawWell.getCell().count / eatingDrawWell.owners.size();
        throw new EventException(EventKind.ACTION_EAT, Math.min(canEat, M - E));
    }

    private void z4() throws EventException {
        throw new EventException();
    }

    private void z5() throws EventException {
        final float neededToEat = M * C_norm - E;
        throw new EventException(EventKind.ACTION_EAT, neededToEat);
    }

    private void z6() throws EventException {
        throw new EventException(EventKind.ACTION_MOVE_ATTACK, foundDestination);
    }

    private void z7() throws EventException {
        throw new EventException(EventKind.ACTION_ATTACK, foundDestination);
    }

    private boolean isGrowingPeriod() {
        return herder.piglets.size() < C_growth_barrier;
    }

    private DrawWell findClosestDrawWell() {
        herder.lockRead();

        DrawWell found = null;
        float minWeight = Herder.INF;
        
        for (DrawWell dw : herder.drawWells) {
            GladeCell cell = dw.getCell();
            
            // first draw well should be good to born
            if ((cell.count + C_steps_to_born * cell.growth) / (dw.owners.size() + 1) < (M - E) / C_steps_to_born) {
                continue;
            }

            // calc draw well weight
            final float dist = distance(location, dw.pos);

            if (dist > C_max_steps_to_dw * S0) {
                continue;
            }

            float weight = dist - (cell.count + cell.growth * C_growt_weight) * C_energy_weight / dw.owners.size();
            if (weight < minWeight) {
                minWeight = weight;
                found = dw;
            }
        }

        herder.unlockRead();

        return found;
    }

    private boolean checkDistanceConstraints(Location next) {
        // distance from current position
        final float dist = distance(location, next);

        final float turnCount = (float)Math.ceil(dist / S);

        // energy at next (consider to move cost and mass-live cost)
        final float E_at_next = E - dist * Constants.K_movecost
                - turnCount * Constants.K_masscost * M;
        
        return E_at_next > (herder.MAX_TURNS - herder.getTurn() - turnCount) * M * Constants.K_masscost + M * C_crit;
    }

    private Location findClosestUnexploredCell() {
        if (herder.everythingExplored) {
            return null;
        }

        herder.lockRead();

        Location found = null;
        float minDist = Herder.INF;

        boolean hasUnexploredColumnt = false;
        for (int i=0; i<herder.glade.length; ++i) {
            if (herder.exploredColumn[i]) {
                continue;
            }
            hasUnexploredColumnt = true;

            boolean hasUnexplored = false;
            for (int j=0; j<herder.glade[i].length; ++j) {
                if (!herder.glade[i][j].explored) {
                    hasUnexplored = true;
                }

                // check if cell hasn't explored
                if (!herder.glade[i][j].explored && herder.glade[i][j].shouldBeExplored != herder.getTurn()) {
                    Location curPos = new Location(i, j);
                    float dist = distance(location, curPos);
                    if (dist < minDist) {
                        found = curPos;
                        minDist = dist;
                    }
                }
            }

            if (!hasUnexplored) {
                herder.unlockRead();
                herder.lockWrite();
                herder.exploredColumn[i] = true;
                herder.unlockWrite();
                herder.lockRead();
            }
        }

        herder.unlockRead();
        
        // mark glade as "should ne explored"
        if (found != null) {
            herder.lockWrite();
            herder.glade[found.getX()][found.getY()].shouldBeExplored = herder.getTurn();
            herder.unlockWrite();
        }

        if (!hasUnexploredColumnt) {
            herder.lockWrite();
            herder.everythingExplored = true;
            herder.unlockWrite();
        }

        return found;
    }

    private static final int [] EXPLORE_DELTA = new int [] { 0, 1, -1, 2, -2, 3, -3 };
    private Location findExploreStep() {
        herder.lockRead();
        
        final int VARIATION_LENGTH = 3;

        int max = 0;
        Location best = null;

        for (int ED_i : EXPLORE_DELTA) {
            for (int ED_j : EXPLORE_DELTA) {
                // let's calc cells amount that will be explored
                int unexplored = 0;

                for (int di = 0; di < VARIATION_LENGTH; ++di) {
                    for (int dj = 0; dj < VARIATION_LENGTH; ++dj) {
                        int x = normalizeX(location.getX() + ED_i + EXPLORE_DELTA[di]);
                        int y = normalizeY(location.getY() + ED_j + EXPLORE_DELTA[dj]);

                        if (!herder.glade[x][y].explored && herder.glade[x][y].shouldBeExplored != herder.getTurn()) {
                            ++unexplored;
                        }
                    }
                }

                if (max < unexplored || max == unexplored && Util.frnd() < 0.2f) {
                    max = unexplored;
                    best = new Location(
                            normalizeX(location.getX() + ED_i),
                            normalizeY(location.getY() + ED_j));
                }
            }
        }

        herder.unlockRead();
        
        if (best != null) {
            herder.lockWrite();

            // mark cells that it will be explored
            for (int di = 0; di < VARIATION_LENGTH; ++di) {
                for (int dj = 0; dj < VARIATION_LENGTH; ++dj) {
                    int x = normalizeX(best.getX() + EXPLORE_DELTA[di]);
                    int y = normalizeY(best.getY() + EXPLORE_DELTA[dj]);

                    herder.glade[x][y].shouldBeExplored = herder.getTurn();
                }
            }

            herder.unlockWrite();
        }
        
        return best;
    }

    private Location findClosestEnemy() {
        if (herder.totalMass < C_fight_barrier) {
            return null;
        }

        herder.lockRead();

        float minDist = Herder.INF;
        Location found = null;

        for (int ED_i : EXPLORE_DELTA) {
            for (int ED_j : EXPLORE_DELTA) {
                final int x = normalizeX(location.getX() + ED_i);
                final int y = normalizeY(location.getY() + ED_j);

                final Location pos = new Location(x, y); 
                final float dist = distance(location, pos);
                if (herder.glade[x][y].enemy >= herder.getTurn() - C_enemy_history && dist < minDist) {
                    minDist = dist;
                    found = pos;
                }
            }
        }

        herder.unlockRead();
        return found;
    }

    private static final int [] FEEDING_DELTA = new int [] { 0, 1, -1, 2, -2, 3, -3, 4, -4, 5, -5, 6, -6 };
    private Location findClosestFeeding() {
        herder.lockRead();

        Location best = null;
        float maxCount = Herder.EPS;

        for (int FD_i : FEEDING_DELTA) {
            for (int FD_j : FEEDING_DELTA) {
                int x = normalizeX(location.getX() + FD_i);
                int y = normalizeY(location.getY() + FD_j);

                GladeCell cell = herder.glade[x][y];

                if (cell.count > maxCount && cell.growth < Herder.EPS) {
                    maxCount = cell.count;
                    best = new Location(x, y);
                }
            }
        }

        herder.unlockRead();

        return best;
    }

    private boolean isCreated = false;
    public BeingParams getParams() {
        if (!isCreated || isGrowingPeriod()) {
            return new BeingParams(M0, S0);
        }
        return new BeingParams(M * Constants.K_maxbornvariation, S0);
    }

    public void reinit(UserGameInfo userGameInfo) {
        super.reinit(userGameInfo);
    }

    private static final Herder herder = new Herder();
    protected SuperHerder getHerder() {
        return herder;
    }

    public void processEvent(BeingInterface bi, Event event) {
        super.processEvent(bi, event);

        switch (event.kind()) {
            case BEING_BORN:
                isCreated = true;
                break;
        }
    }

    public void log(String s) {
        if (ID > 30) {
            return ;
        }

        if (beingInterface != null) {
//            beingInterface.log(this, "(" + E + "/" + M + ") " + s);
        }
    }
}
